#! /usr/bin/gawk -f
# ######################################################################### #
#                                                                           #
# File:     Tokenize.awk                                                    #
# Authors:  Brigitte Bigi                                                   #
# Date:     August, 2008                                                    #
# Licence:  GPL                                                             #
# Brief:    Lexical tokenizer, using rules                                  #
# Help:	    Tokenize.awk -h                                                 #
#                                                                           #
# ######################################################################### #


# ######################################################################### #
#                                                                           #
#  Function FCT_LoadVocab                                                   #
#                                                                           #
#  Description: Load a vocabulary from a file                               #
#                                                                           #
#  Parameters:                                                              #
#       file is the file name of the vocabulary                             #
#                                                                           #
# ######################################################################### #
function FCT_LoadVocab (file)
{
	_ind = 0
	while ( getline < file )
	{
		TAB_Voc[$0]=_ind
		_ind++
	}
	return _ind
}


# ######################################################################### #
#                                                                           #
#  Function FCT_IsOOV                                                       #
#                                                                           #
#  Return value:                                                            #
#       boolean value: 1 if word is in TAB_Voc, 0 else                      #
#                                                                           #
# ######################################################################### #
function FCT_IsOOV (word)
{
	if (TAB_Voc[word] != "") return 0
	return 1
}


# ######################################################################### #
#                                                                           #
# Function Name: FCT_SplitPunct                                             #
#                                                                           #
# Description:                                                              #
#     Add spaces before and after some specific punctuations                #
#                                                                           #
# ######################################################################### #
function FCT_SplitPunct ( str )
{
	# Cas du .
	str = gensub(/([a-z0-9")])\./, "\\1_.", "g", str)
	if (length(str) > 3)
		str = gensub(/([A-Z])\./, "\\1_.", "g", str)

	# Eclate des guillemets, les parenthses, les # et les virgules
	gsub(/,/ "_,_", str)
	gsub(/[(]/, "_(_", str)
	gsub(/[)]/, "_)_", str)
	gsub(/#/, "_#_", str)
	gsub(/,/, "_,_", str)

	# Eclate les ? ! : "
	gsub(/!/, "_!_", str)
	gsub(/?/, "_?_", str)
	gsub(/:/, "_:_", str)
	gsub(/"/, "_\"_", str)
	return str
}


# ######################################################################### #
#                                                                           #
#  Function Analyse                                                         #
#       Analyze and Tokenize a string which is out-of-vocabulary            #
#                                                                           #
# Parammeters:                                                              #
#       token is the string to analyze                                      #
#       Lst is the resulted tabular                                         #
#                                                                           #
# ######################################################################### #
function Analyse (token, Lst)
{
	# Find a delimiter in the token
	# (cases where the token is systematically tokenized)
	if ( (match (token, /[\",?!;:\(\)\&\/\`\%\*\[\]\\]/) != 0) || 
	     (match (token, /[ [:alnum:]][\]/) != 0) )
	{
		token = gensub (/([\",?!;:\(\)\&\/\`\%\*\[\]\\])/, " \\1 ", "g", token)
		token = gensub (/([ [:alnum:]])([\])/, "\\1 \\2 ", "g", token)
		return split(token, Lst, " ")
	}
	else
	{
		if (lang == "EN")
		{
			if ( match (token, /['s]/ ) != 0 )
			{
				# Split just before the '
				sub(/'/, " '", token)
				_nb=split(token, Lst, " ")
				return _nb
			}
		}

		if ( match (token, /[']/ ) != 0 )
		{
			# Split just after the '
			sub(/'/, "' ", token)
			_nb=split(token, Lst, " ")
			return _nb
		}
		else
		{
			if ( match (token, /[\.]/ ) != 0)
			{
				# Special case: ... at the end of a word
				if ( length(token) > 3 &&  token ~ /\.\.\.$/ ) 
				{
					substr(token, 0, length(token)-3)
					Lst[1] = token
					Lst[2] = "..."
					return 2
				}
				else
				{
					if  ( ( length(token) < 3 ) || ( index(token,".") < 3 ) )
					{
						Lst[1] = token
						return 1
					}	
					else
					{
						_nbtmp = split (token, _LstTmp, ".") 
						_j = 1
						for (_i=1; _i<_nbtmp; _i++)
						{
							Lst[_j] = _LstTmp[_i]
							_j++
							Lst[_j] = "."
							_j++
						}
						Lst[_j] = _LstTmp[_i]
						delete _LstTmp
						return _j
					}
				}

			}
			else
			{
				if ( token ~ /[-]/ )
				{
					# Write mot1
					_ind   = index(token, "-")
					Lst[1] = substr(token, 1, _ind-1)

					# Find -mot2 in the vocabulary
					_tok = substr(token, _ind, length(token)-_ind+1)
					if (FCT_IsOOV(_tok) == 0)
					{
						# Write -mot2
						Lst[2] = _tok
						return 2
					}
					else
					{
						# Split -mot2
						token = substr(_tok, 2, length(token)-1);
						Lst[2] = "-"
						Lst[3] = token
						return 3
					}
				}
				else
				{
					# Unrecognized situation: write without any change
					Lst[1] = token
					return 1
				}
			}
		}
	}
}


# ######################################################################### #
#                                                                           #
# Function Name: FCT_Usage                                                  #
#                                                                           #
# Description: Print the usage and exit the program                         #
#                                                                           #
# ######################################################################### #
function FCT_Usage ()
{
        printf "Usage: Tokenize.awk [options]\n\n"
        printf "Where options include:\n"
        printf "  -o filename    write the tokenized corpus to this file (default: stdout)\n"
	printf "  -l lang        EN or anything else!\n"
        printf "  -V filename    read vocabulary from this file\n"
        printf "  -h             print this help\n"
	printf "\n"
	printf " input text:  stdin\n"
	printf " output text: stdout (except if -o)\n"
        exit
}


# ######################################################################### #
#                                                                           #
# Function Name: FCT_CommandLine                                            #
#                                                                           #
# Description: Gets arguments from the command line                         #
#                                                                           #
# ######################################################################### #
function FCT_CommandLine(argc, argv)
{
        for (i=1; i<argc; i++)
        {
                if (argv[i] == "-h")
                        FCT_Usage()
                else
                if (argv[i] == "-o")
                {
                        Output = argv[i+1]
                        i++
                }
                else
                if (argv[i] == "-l")
                {
                        lang = argv[i+1]
                        i++
                }
                else
                if (argv[i] == "-V")
                {
                        vocabsize = FCT_LoadVocab(argv[i+1])
                        i++
                }
                else
                if (argv[i] ~ /^\-/)
                {
                        printf "invalid option -- %s\n", argv[i] > "/dev/stderr"
                        FCT_Usage()
                }
        }
}


# ############################# BEGIN MAIN ################################ #

BEGIN{
        Output      = "/dev/stdout"    # output file name
        vocabsize   = 0                # vocabulary size
	lang        = "FR"
        FCT_CommandLine(ARGC, ARGV)    # extract parameters
        delete ARGV
        ARGC=1

	# Index des elements d'une phrase
	idxstr=0
}


/<str>/{
        # Get the element
        s = index($0, ">")				# start: 1st >
        e = index(substr($0, s+1, length($0)), "<")	# end: 1st < after start
        w = substr($0, s+1, e-1)			# word: $0 from start to end
        gsub(/ /, "", w)				# remove spaces

	# Pas de tokenize, on laisse tel quel
	# 	- mot du lexique
	#	- tag
	#	- caractere de ponctuation
	#	- adresse internet
	if ( FCT_IsOOV(w) == 0  || 
	     w ~ /<?orig>/ || 
	     match(w, /^[\.\-'\",?!;:\(\)&\/\`%\*\[\]\\\_]$/) != 0 ||
	     match (w, /^http/) != 0 || match (w, /^www/) != 0 )
	{
		print "<wd>\n<str>", w , "</str>\n</wd>" > Output
	}
	else
	{
		# On s'occupe d'eclater les caracteres de ponctuation
		# sauf dans le cas d'une adresse internet
		if ( match (w, /^http/) == 0 && match (w, /^www/) == 0 )
			w = FCT_SplitPunct(w)
		gsub(/__/, "_",w)
		gsub(/^_/, "", w)
		gsub(/_$/, "", w)
		nb1 = split(w, Lst_Elt1, "_")

		for (i=1; i<=nb1; i++)
		{
			nb2 = Analyse(Lst_Elt1[i], Lst_Elt2)
			if (nb2 == 1)
			{
				# supprime les espaces
				gsub(/ /, "", Lst_Elt1[i])
				# ecrit l'element
				if (length(Lst_Elt1[i]) > 0)
					print "<wd>\n<str>", Lst_Elt1[i], "</str>\n</wd>" > Output
			}
			else
			{		
				# Double analyse de la chaine, 
				# au cas ou il faudrait un double tokenize
				for (j=1; j<=nb2; j++)
				{
					nb3 = Analyse(Lst_Elt2[j], Lst_Elt3)
					for (k=1; k<=nb3; k++) 
					{
						# supprime les espaces
						gsub(/ /, "", Lst_Elt3[k])
						# ecrit l'element
						if (length(Lst_Elt3[k]) > 0)
							print "<wd>\n<str>", Lst_Elt3[k] ,"</str>\n</wd>" > Output
					}
					delete Lst_Elt3
				}
			}
			delete Lst_Elt2
		}
		delete Lst_Elt1
	}
	next
}

/<[\/]?wd>/ { next }

{
	print > Output
}
# ############################## END MAIN ################################# #


